
/*
 * MGL graphics library
 * Copyright (C) 1999-2000, ing. Marek Gayer

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.

 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <mgl.h>

mglFont *mglNewFont (void)
{
  mglFont *font;
  int i;

  font = _malloc (sizeof (mglFont));
  font->texture = NULL;
  font->w = font->h = 0;
  font->spacingX = font->spacingY = 0;
  for (i = 0; i < 256; i++)
  {
    font->chars [i].x = font->chars [i].y = -1;
  }

  return font;
}

void mglLoadFontDefinitionTextFile (mglFont *font, char *textFileName)
{
  FILE *f;
  char buf [80];
  mglXY r;
  char c;
  int code;
  int line;
  char name [80];
  char s [1024];

  if ( ( f = fopen( textFileName, "rt" ) ) == NULL )
  {
    sprintf (s, "font definition text file not found (%s).", textFileName);
		mglError (MGL_ERROR_FONT_DEFINITION_FILE, s);
	}

  line = 0;
  while (fgets (buf, 80, f) != NULL)
  {
    if (buf [0] == '\n')
      continue;

    if (line == 0)
      sscanf (buf, "%s %i %i %i %i", name, &font->w, &font->h, &font->spacingX, &font->spacingY);
    else
    {
      c = 0; code = -1;
      sscanf (buf, "%i %i %i %i %c (%i)", &r.x, &r.y, &r.width, &r.height, &c, &code);
      if (c)
      {
        if (r.width == 0)
          r.width = font->w;
        if (r.height == 0)
          r.height = font->h;

        if ((c == '?') && (code != -1))
          c = code;

        font->chars [(unsigned char)c] = r;
      }
    }
    line ++;
  }

  fclose (f);
}


mglFont *mglLoadFont (char *textFileName, char *imageFileName)
{
  mglImage *i;
  mglFont *font;

  font = mglNewFont ();
  mglLoadFontDefinitionTextFile (font, textFileName);

  i = mglLoadImage (imageFileName, 2);

  font->texture = mglMakeTexture (i, GL_LINEAR_MIPMAP_NEAREST);
  mglDeleteImage (i);

  mglNote ("font created successfully.\n");

  return font;
}

void mglDeleteFont (mglFont *font)
{
  mglDeleteTexture (font->texture);

  _free (font);
}

static float defRotation;


int mglPutc (mglFont *font, char c, float x, float y)
{
  int fromX;
  int fromY;
  int width;
  int height;
  unsigned char cc = c;

  if (font->chars [cc].x == -1)
  {
    if (islower (cc))
      cc = toupper (cc);
    else
    if (isupper (cc))
      cc = tolower (cc);
  }

  if (font->chars [(unsigned char)cc].x != -1)
  {
    defRotation = mglRotationAngle;
    mglRotationAngle = 0;

    fromX = font->chars [cc].x;
    fromY = font->chars [cc].y;
    width = font->chars [cc].width;
    height = font->chars [cc].height;

    mglDrawSubTexture (font->texture, x, y, fromX, fromY, width, height);

    mglRotationAngle = defRotation;

    return width;
  }
  else
    return font->w;
}

int mglFontWrap;
int mglFontWordWrap;

static int wordLen (char *s)
{
  int len;

  len = 0;
  while (*s != '\n' && *s != ' ' && *s != 0)
  {
    if (*s == '\a')
    {
      s++;
      if ((*s > '9') || (*s < '0'))
        len++;
    }
    else
      if (*s != '\t')
        len++;
    s++;
  }

  return len;
}

void mglPuts (mglFont *font, char *s, float x, float y)
{
  int i;
  float oldX;
  float oldY;
  char c;
  float a;
  int wrap;
  int len;
  mglColor rect [4];
  mglColor col;
  int wordWrap;

  wordWrap = mglFontWordWrap;

  wrap = 0;

  a = mglTextureColor [0].a;

  i = 0;
  oldX = x;
  oldY = y;

  rect [0] = mglTextureColor [0];
  rect [1] = mglTextureColor [1];
  rect [2] = mglTextureColor [2];
  rect [3] = mglTextureColor [3];

  while (c = s [i], c != 0)
  {
    if ((wrap > mglFontWrap-1) && mglFontWrap && !mglFontWordWrap)
    {
      x = oldX;
      y += (font->h + font->spacingY) * mglTextureZoomY;
      wrap = 0;
      wordWrap = mglFontWordWrap;
      continue;
    }
    switch (c)
    {
      case '\t':
        i = i;
        break;
      case '\n':
        x = oldX;
        y += (font->h + font->spacingY) * mglTextureZoomY;
        wrap = 0;
        wordWrap = mglFontWordWrap;
        break;
      case '\a':
      {
        i++;
        c = s [i];
        if (c == '0')
        {
          mglTextureColor [0] = rect [0];
          mglTextureColor [1] = rect [1];
          mglTextureColor [2] = rect [2];
          mglTextureColor [3] = rect [3];
        }
        else
        {
          if ((c >= '1') && (c <= '9'))
          {
            col = *mglColorBase [c-'1'];
            col.a = a;
            mglSetTextureColors (col);
          }
        }
        break;
      }
      default:
        wrap++;
        if (mglFontWrap && wordWrap)
        {
          len = wordLen (s+i);
          if ( ((wrap + len) > mglFontWrap+1) && len && c != ' ')
          {
//            if (len > mglFontWrap+1)
              //wordWrap = 0;
//            else
            {
              wrap = 1;
              x = oldX;
              y += (font->spacingY + font->h) * mglTextureZoomY;
            }
          }
        }
        x += (mglPutc (font, c, x, y) + font->spacingX) * mglTextureZoomX;
        if (c == ' ')
          wordWrap = mglFontWordWrap;
        break;
    }
    i++;
  }
}

